
      subroutine tabular (pedig,n,m,priorupg,nupgcur,inb)
c Calculates inbreeding coefficients using tabular method.
c pedig - list of pedigrees of n animals, ordered by birth date
c priorupg - prior inbreeding coeeficients for nupg unknown parent groups
c            (useful if unknown parents are assumed inbred)
c nupgcur - number of unknown parent groups, or 0 if no change
c inb - calculated inbreeding coefficients
c
      integer n,m,nupg,nupgcur,maxped,msglev/0/
c maxped = 100 + maximum number of ancestors for any animal
      parameter (maxped=100+420)
      integer pedig(m,n),i,j
      real priorupg(nupgcur),inb(n),tabmat(maxped,maxped)
      save tabmat,nupg

c load unknown parent group inbreeding coefficients in the 0 row and
c column of maxped
      if (nupgcur.gt.0) then
          if (nupgcur.gt.maxped) then
             print*,'TABULAR: maxped should be >> ',nupgcur
             stop
          endif
          nupg=nupgcur
          do i=1,nupg
             do j=1,nupg
               tabmat(i,j)=2*priorupg(max(i,j))
             enddo
             do j=1+nupg,maxped
               tabmat(i,j)=2*priorupg(i)
               tabmat(j,i)=tabmat(i,j)
             enddo
          enddo
      endif


      if (n+nupg .gt. maxped) then
         print*,'MAXPED in TABULAR should be at least ',n+nupg
         stop
      endif

c set diagonals
      do i=nupg+1,n+nupg
           tabmat(i,i)=1
      enddo

c actual computations

      do i=1,n
c          diagonal element first
         tabmat(i+nupg,i+nupg)=tabmat(i+nupg,i+nupg)+
     +               .5*tabmat(pedig(1,i)+nupg,pedig(2,i)+nupg)
c          off diagonals
         do j=i+1,n
              tabmat(i+nupg,j+nupg)=.5*
     +             (tabmat(i+nupg,pedig(1,j)+nupg)+
     +              tabmat(i+nupg,pedig(2,j)+nupg))
              tabmat(j+nupg,i+nupg)=tabmat(i+nupg,j+nupg)
         enddo
      enddo

c
c copy diagonals
      do i=1,n
         inb(i)=tabmat(i+nupg,i+nupg)-1.
      enddo

c print diagnostics if msglev>0
      if (msglev .gt.0) then
         print*,'pedigree'
         do i=1,n
            print '(2i10)',pedig(1,i),pedig(2,i)
         enddo
         if (nupgcur.ne.0) print '('' prior inbreeding''100(16f5.2/))',
     +                  (priorupg(i),i=1,nupg)
         print*,'A'
         do i=nupg+1,n+nupg
            print '(16f5.2)',(tabmat(i,j),j=nupg+1,n+nupg)
         enddo
      endif
      end

      subroutine rentab(x,m,n)
c renumber pedigree table for tabular method so that the oldest animal is 1
      integer m,n,x(m,n),i,j,k,an
c
c renumbered number are initially <0
c
      do i=1,n
         an=x(3,i)
         do j=1,2
            do k=1,n
               if (x(j,k).eq.an) x(j,k)=-i
            enddo
         enddo
      enddo

c     negate all numbers, and alert if some numbers not renumbered
      do i=1,2
         do j=1,n
            if (x(i,j).le.0) then
               x(i,j)=-x(i,j)
              else
               print*,'HAINB:number not renumbered in rentab:',x(i,j)
               stop
            endif
         enddo
      enddo
      end

      function addr(animal,init_an,init_group,nherd,curherd,hpoint)
c determines new address of animal given its local/global status
        integer addr,animal,init_an,init_group,nherd,curherd,
     +          hpoint(nherd)
c
        if (animal .gt.init_group) then
           addr=0
        elseif (animal .gt.init_an) then
           addr=animal-init_an+hpoint(nherd)
        elseif (animal.le.0) then
            addr=0
        else
            addr=animal+hpoint(curherd)
        endif
        end

      subroutine printmat(a,x,m,n,column)
c prints text a, and then integer matrix x by columns if column=1, or by
c rows otherwise
c
      integer n,m,column,x(m,n),i,j
      character a*(*)

      print '(1x,a)',a
      if (column.eq.1) then
         do j=1,n
            print '(10i8)',(x(i,j),i=1,m)
         enddo
        else
         do i=1,m
            print '(12i6)',(x(i,j),j=1,n)
         enddo
      endif
      end

