c------------------------------------------------------------------------
c Program JAADOM version 2.2 - works with the dominance effect
c  Copyright Ignacy Misztal & University of Georgia
c------------------------------------------------------------------------
c  1/18/95 - 2/14/96
c
c  Research use of this program is permitted if acknowledged in publications.
c
c  Parameters
c     mx - total number of equations
c
c  Vectors
c     acc(mx)  - adjusted right-hand sides
c     diag(mx) - diagonals of the system
c     sol(mx)  - solutions
c     dsol(mx) - old solutions
c
c     sola(mx) - alternate sol and sola for implementing Method R
c     dsola(mx)
c-----------------------------------------------------------------------      
c
      integer mx,irew,iopen,iwrite,iread
      parameter (mx=2 100 000)
      parameter(irew=1,iopen=2,iwrite=3,iread=4)
      real  diag(mx),sol(mx),dsol(mx),sola(mx),dsola(mx),
     +      var(20),buf(50),val(20),qq,rel,rel1,y,type,add,xx,
     +      ran1,sel,beta
      real*8 acc(mx)
      real*8 uau1,uau,adj,adj0
      integer ifact(20),ifact1(20),ilev(20),ilev1(20),ipos(20),ifil(9),
     +        nunin,nunout,nooffa,imax,iposmx,isize,irelat,lastan,
     +        i,ii,j,j1,j2,itr,jj,k,k1,l,l1,n,n1,nr,seed,iran
      logical ireg(20),even,extrap_r
      character*40 nampar,namdat,namrel,namout
c
c storage for dominance and method r
      integer nrel,ndom,msglev/1/
      real w(9)/1.0,4*-.5,4*.25/,wa(3)/1.,2*-.5/,r(0:255),corr(20)

c definition of an even function
      even(i)=mod(i,2).eq.0

c compute residuals for different classes of missing s-d combinations
      call creatres(r)

      print*,'fractions of records selected (.01-.99),random seed, beta'
      read*,sel,seed,beta
      print*,'beta=',beta

c the parameters can be read either from the console or from a file
      write(*,*)'name of a file with parameters, con - if console:'     
      read(*,3)nampar
3     format(a40)
      if (nampar.ne.'con') then
          nunin=1
          nunout=20
          open(nunin,file=nampar)
          open(nunout,file='null')
        else
          nunin=5
          nunout=6
      endif
      write(nunout,*)'data name:'
      read(nunin,3)namdat
      write(nunout,*)'name of file with animal relationships:'
      read(nunin,3)namrel
      write(nunout,*)'name of output file:'
      read(nunin,3)namout
      open(2,file=namdat)
      open(3,file=namrel)
      open(7,file=namout)
      open(10,file='dom'//namrel)
5     write(nunout,*) 'no of factors, trait pos., rel. factor,'
     * ,'iterations?:'  
      read(nunin,*)nooffa,itr,rel,imax                                      
      write(nunout,*)' for each factor give position, no of levels,'   
      write(nunout,*)' and variance ratio, 0 if fixed, negative',
     * ' if relat. matrix'
      write(nunout,*)'  no of levels < 0 denotes covariable'
      write(nunout,*)
      iposmx=itr                                                        
      isize=0                                                           
      irelat=0                                                          
      do i=1,nooffa                                                  
        read(nunin,*)ipos(i),ilev(i),var(i)                                   
c number of levels <0  denotes covariable
        if (ilev(i).lt.0) then
              ilev(i)=1
              ireg(i)=.true.
           else
              ireg(i)=.false.
        endif
        if (var(i).lt.0) then                                             
          if (irelat.eq.1) then                                        
            print*,'only one relat. matrix allowed, type all again' 
            stop                                                            
         else                                                              
          irelat=1                                                          
        endif                                                             
        endif
        iposmx=max(iposmx,ipos(i))                                        
        ilev1(i)=isize                                                    
      isize=isize+ilev(i)
      enddo
c
cdom read dominance variables
      write(nunout,*) 'dominance variables?'
      read(nunin,*)nrel,ndom
      print*,'nrel,ndom',nrel,ndom
c
*                                                                       
      if (isize.gt.mx) then
         print*,'too large model, increase parameter mx'
         stop
      endif
      write(*,*)'iteration no.  convergence criterion   last solution'
      write(7,*)'iteration no.  convergence criterion   last solution'
* rewrite the data to an unformatted file to gain on speed              
      call iof(iopen,ifil,n,4)
      call iof1(iopen,ifil,n,8)
      call iof2(iopen,ifil,n,9)

      n=0                                                               
22    read(2,*,end=24)(buf(i),i=1,iposmx)                               
      n=n+1                                                             
      do  i=1,nooffa                                                  
        ifil(i)= int(buf(ipos(i)))
      enddo
      ifil(nooffa+1)=buf(itr)                                         
      n1=nooffa+1
      call iof(iwrite,ifil,n1,4)     
      goto 22                                                           
*                                                                       
24    if (msglev.ge.1) print*,n,' records'
      nr=0
      if (irelat.eq.1) then
        lastan=0
27      read(3,*,end=26)(ifact(i),i=1,4)
        lastan=max(ifact(1),lastan)
        nr=nr+1
        call iof1(iwrite,ifact,4,8)
        goto 27
      endif

cdom copy also the dominance relationship matrix
26    if (msglev.ge.1) print*,nr,' additive relationships'
      ndom=0
32    read(10,*,end=33)(ifact(j),j=1,10)
      ndom=ndom+1
      call iof2(iwrite,ifact,10,9)
c     print '(10i7)',(ifact(j),j=1,10)
      goto 32

33    print*,ndom,' dominance relationships'

      do i=1,mx
        dsol(i)=0
        sol(i)=0
        sola(i)=0
        dsola(i)=0
      enddo
      i=1                                                               
*                                                                       
25    adj=0
      adj0=0
      do j=1,isize
        acc(j)=0
        diag(j)=0
      enddo

* do one round of gauss-seidel (jacobi if relationship matrix)
      do  jj=1,nooffa           ! main loop
        uau=0
        uau1=0
        iran=-seed
        xx=ran1(iran)
        call iof(irew,ifact,0,4)
        l=0
        l1=0
        do  ii=1,n
          call iof(iread,ifact1,nooffa+1,4)
          l=l+1
          if (ran1(iran).gt.sel .and. even(i)) goto 30    ! selection of records for R
c          if (l.ge.5 .and. even(i)) goto 30
          l1=l1+1
          y=float(ifact1(n1))
          do j=1,nooffa
           if (ireg(j)) then
                val(j)=ifact1(j)
                ifact1(j)=1
            else
              if (ifact1(j).le.ilev(j) .and. ifact1(j).ge.1) then
                  val(j)=1
                 else
                  val(j)=0
              endif
           endif
           ifact(j)=ifact1(j)+ilev1(j)
          enddo
          j1=ifact(jj)
          diag(ifact(jj))=diag(ifact(jj))+val(jj)*val(jj)
          acc(ifact(jj))=acc(ifact(jj))-y*val(jj)
        do  k=1,nooffa
            k1=ifact(k)
            if (k1 .ne. j1) acc(ifact(jj))=acc(ifact(jj))+
     +                         sol(k1)*val(jj)*val(k)
        enddo
30    continue
      enddo
 
      if (i.le.5 .and. jj.eq.1) then
         print '('' selected'',i9,'' records out of'',i9)',l1,l
      endif
    
      if (msglev.ge.2) then
        print*,'jj',jj
         print '(''sol'',10f7.2)',(sol(j),j=1,isize)
         print '(''dia'',10f7.2)',(diag(j),j=1,isize)
         print '(''acc'',10f7.2)',(acc(j),j=1,isize)
         print*
      endif

* solve each factor separately, only fixed factors during the first
* round
40    j1=ilev1(jj)+1
      j2=ilev(jj)+ilev1(jj)

c===========================================================================
* add variance ratio to diagonals of random effects

       if (jj.eq.nooffa.and.var(jj).gt.0) then
c dom - a dominance effect
         if (msglev.ge.3) print '(''d1d'',10f7.2)',(diag(j),j=1,isize)
c--------- first add diagonals to effects without dom pedigree line
          do j=ndom+1,ilev(jj)
             k=ilev1(jj)+j
             diag(k)=diag(k)+var(jj)
             uau1=uau1+var(jj)*sol(k)*sola(k)
             uau =uau +var(jj)*sol(k)*sol(k)
          enddo
          if (msglev.ge.3) print '(''d2d'',10f7.2/)',(diag(j),j=1,isize)
c
c---------now the real dominance matrix
          call iof2(irew,ifact,0,9)
          do j=1,ndom
             call iof2(iread,ifact,10,9)
             if (r(ifact(10)).eq.0) then
                print*,'weight=0 for dominance line',j,';',ifact(1)
                stop
             endif
             qq=var(jj)*r(ifact(10))
             if (msglev.ge.4) print*,var(jj),ifact(10),r(ifact(10)),qq
             do k=1,9
                k1=ifact(k)
                if (k1.ne.0) then
                   k1=k1+ilev1(jj)
                   diag(k1)=diag(k1)+qq*w(k)*w(k)
                   do l=k+1,9
                      l1=ifact(l)
                      if (l1.ne.0) then
                         l1=l1+ilev1(jj)
                         acc(k1)=acc(k1)+qq*sol(l1)*w(k)*w(l)
                         acc(l1)=acc(l1)+qq*sol(k1)*w(k)*w(l)
                      endif
                   enddo
                   do l=1,9
                      l1=ifact(l)
                      if (l1.ne.0) then
                         l1=l1+ilev1(jj)
                         uau1=uau1+qq*sol(l1)*sola(k1)*w(k)*w(l)
                         uau =uau +qq*sol(l1)*sol(k1) *w(k)*w(l)
                      endif
                   enddo
                endif
             enddo
           enddo

          if (uau.ne.0 ) then
              corr(jj)=uau1/uau
              if (msglev.ge.2) print*,'jj,uau1,uau',jj,uau1,uau
          endif
          if (msglev.ge.3) print '(''d3d'',10f7.2)',(diag(j),j=1,isize)
       endif
c

       if (var(jj).gt.0.and. jj.ne.nooffa) then
           do  k=j1,j2
             diag(k)=diag(k)+var(jj)
             uau1=uau1+sol(k)*sola(k)
             uau =uau +sol(k)*sol(k)
           enddo
          if (uau.ne.0 ) then
              corr(jj)=uau1/uau
              if (msglev.ge.2) print*,'jj,uau1,uau',jj,uau1,uau
          endif
       endif

       if (var(jj).lt.0) then
* add additive relationship matrix if such exists
         call iof1(irew,ifact,0,8)
         do  k=1,nr
           call iof1(iread,ifact,4,8)
           do l=1,3
              ifact(l)=ifact(l)+ilev1(jj)
           enddo
c---------- ifact(4) contains UPG code, or a function of UPG with inbreeding
           if (ifact(4).le.3) then
                 type=-var(jj)*4/(1+ifact(4))
               else
                 type=-var(jj)*ifact(4)/1000.
           endif
           do l=1,3
              xx=wa(l)*type
              diag(ifact(l))=diag(ifact(l))+xx*wa(l)
              do j=l+1,3
                 acc(ifact(l))=acc(ifact(l))+xx*wa(j)*sol(ifact(j))
                 acc(ifact(j))=acc(ifact(j))+xx*wa(j)*sol(ifact(l))
              enddo
               do j=1,3
                uau1=uau1+sola(ifact(l))*sol(ifact(j))*wa(l)*wa(j)*type
                uau=uau+sol(ifact(l))*sol(ifact(j))*wa(l)*wa(j)*type
              enddo
           enddo
        enddo
c        if (i.eq.1) print*,' temporary fix to null UPGs'
c        do j=j1+lastan,j2
c           acc(j)=0
c        enddo
        if (uau.ne.0 ) then
            corr(jj)=uau1/uau
            if (msglev.ge.2) print*,'jj,uau1,uau',jj,uau1,uau
        endif
      endif

* solve
      do  k=j1,j2
      rel1=0
        if (var(jj).lt.0) rel1=rel
        if (diag(k).ne.0) then
          add=-acc(k)/diag(k)-sol(k)
          adj=adj+add*add
          xx=sol(k)
          if (jj.eq.nooffa .and. var(jj).gt.0) then
c----------dominance effect
              sol(k)=sol(k)+rel*(sol(k)-dsol(k))+beta*add
            else
c----------other effects
              sol(k)=sol(k)+rel1*(sol(k)-dsol(k))+add
          endif
          dsol(k)=xx
          adj0=adj0+sol(k)*sol(k)
        endif
      enddo
      
      enddo     !end of big loop
* 
      if (msglev.ge.2) then
        print '(''ssol'', 10f7.2)',(sol(j),j=1,isize)
        print '(''sdia'',10f7.2)',(diag(j),j=1,isize)
        print '(''sacc'',10f7.2)',(acc(j),j=1,isize)
        print*
      endif

      adj=adj/adj0
      write(7,'(i4,2g10.2)')i,adj,sol(isize)
      if (msglev.gt.1 .or. mod(i,2).eq.1) 
     +                 write(*,'(i4,2g10.2)')i,adj,sol(isize)
c
c Method R, extrapolate if conditions met
      if (even(i)) then
c          print '(26x,''r='',10f9.5)',(corr(j),j=1,nooffa)
           if (extrap_r(nooffa,var,i/2,adj,corr)) then
              call swapv(sol,sola,isize)
              goto 500
           endif
      endif
      i=i+1                                                             
      call swapv(sol,sola,isize)
      call swapv(dsol,dsola,isize)
      if (i.lt.imax ) goto 25                                           

      print '('' how many more iterations, 0-none?, new vars(1=yes)?'')'
      read(*,*)j,k
      if (j.gt.0) then                                                  
         imax=imax+j
         if (k.eq.1) then
           print '(''current ratios:'',9f7.2)',(var(k),k=1,nooffa)
           print*,'effect number and variance'
           read*,k,xx
           var(k)=sign(abs(xx),var(k))
           print '(''new  ratios:   '',9f7.2)',(var(k),k=1,nooffa)
         endif
         goto 25
       elseif (even(i)) then
         call swapv(sol,sola,isize)
         call swapv(dsol,dsola,isize)
      endif                                                             

c write down solutions
500   write(7,*)
      write(7,*) n,' records',isize,' equations'
      do i=1,nooffa
        write(7,*)
        write(7,*) 'solutions of factor',ipos(i),'  var. ratio',var(i)
        if (ilev(i).lt.100) write(7,570)(j,sol(j+ilev1(i)),j=1,ilev(i))
      enddo
570   format(4(3x,i4,':',f11.4))

      open(21,file='sol')
      do i=1,isize
        write(21,'(f8.2)')sol(i)
      enddo
      print*,'solutions were stored in file sol'
      close(4,status='delete')
      close(8,status='delete')
      close(9,status='delete')
      end
 
      subroutine swapv(x,y,n)
c swap real vectors x and y
      integer n,i
      real x(n),y(n),temp
c
      do i=1,n
         temp=x(i)
         x(i)=y(i)
         y(i)=temp
      enddo
      end

